#include "main.h"
#include "filed.h"
#include "sys.h"
#include "flags.h"
#include "log.h"
#include "conf.h"
#include "socket.h"
#include "io.h"
#include "db.h"
#include "shmem.h"
#include "dbq.h"
#include "peer.h"
#include "session.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <malloc.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

struct filed *filed;
int sys_flags;
int debug;

int main(int argc, char *argv[]) {
	pid_t pid;
	FILE *fp;

	sys_flags = 0;
	debug = 0;

	init_filed();

	if(!grow_shmem()) {
		printf("Unable to aquire enough shared memory\n");
		display_shmem();
		exit(1);
	}

	parse_options(argc,argv);
	setup_signals();
	read_cf();

	if(shmem_check()) {
		elog("another filed is already running");
		die();
	}

#ifndef NOFORK
	if((pid = fork())) {
		/* parent process */
		elog("Filed %s forked. (%d)",VERSION,pid);
		if((fp = fopen("/var/run/filed.pid","w"))) {
			fprintf(fp,"%d\n",pid);
			fclose(fp);
		}

		exit(0);
	} else {
#endif

		/* child process */


		if((filed->db_pid = fork())) {
	
			bind_shmem();
			bind_socket();
			wait_for_db();
			if(!(sys_flags & FLAG_UP)) sys_flags += FLAG_UP;
			dlog(DEBUG_THREAD,"main thread ready");
			
			setup_stationd();

			while(sys_flags & FLAG_UP) {
				handle_connections();
				if(not_overloaded()) process_io();
				handle_shmem();
				handle_timing();
				handle_peers();
				dbq_process();
			}

			die();
		} else {
			usleep(10);
			bind_shmem_client();
			wait_for_main();
			if(debug & DEBUG_THREAD) elog("database thread ready");
			handle_db(); /* at some later date we may want */
				     /* to be able to have N database threads */
		}
	
#ifndef NOFORK
	}
#endif
	exit(0);
}

void handle_timing(void) {
	time_t t;

	filed->spin++;
	if(filed->spin > 100) {
		t = time(NULL);
		filed->run = t - filed->last_hb;
		filed->last_hb = t;
		filed->load = get_load();
		filed->spin = 0;
		reconnect_peers();
		scan_filesystems();
		determine_best_peer();
		if(shmem_ping_db_thread()) {
			elog("Database thread has gone away!");
			die();
		}
		dlog(DEBUG_HB,"HB %d %d",filed->load,filed->run);
		shmem_respond_to_filed_pings();
		shmem_handle();
		session_do_timeouts();
	}

	return;
}

void init_filed(void) {
	char *c;
	if(!(filed = (struct filed *) malloc(sizeof(struct filed)))) {
		elog("Unable to malloc memory");
		exit(0);
	}

	filed->con_head = NULL;
	filed->session = NULL;
	
	filed->cf = NULL;
	filed->db = NULL;
	filed->msg = NULL;
	filed->todel = NULL;
	filed->id = 0;
	filed->dbid = 0;
	filed->session_id = 0;
	filed->request_id = 0;
	filed->peer_flags = 0;
	filed->s_count = 0;
	filed->best_peer = 0;
	filed->main_fd = 0;
	sprintf(filed->conf_file,"/etc/filed.conf");
	c = get_hostname();
	if(c)
	strcpy(filed->hostname,c);
	else
	strcpy(filed->hostname,"");
}
	
void parse_options(int argc, char *argv[]) {
	/* parse options, set any sys flags or whatever */
	/* todo: write this */
	int i;

	for(i=1;i<=argc;i++) {
		if(argv[i] && argv[i][0] == '-') {
			switch(argv[i][1]) {
				case '?':
				case 'h':
					usage();
					exit(0);
				case 'v':
				case 'V': 
					version();
					exit(0);
				case 'S':
					display_shmem();	
					exit(0);
				case 'c':
					if(!(sys_flags & FLAG_CONSOLE_WRITE))
						sys_flags += FLAG_CONSOLE_WRITE;
					break;
				case 'd':
					i++;
					if(!argv[i]) {
						printf("error: -d requres an argument\n");
						usage();
						exit(0);
					}
					debug = atoi(argv[i]);
					break;
	
				case 'f':
					i++;
					if(!argv[i]) {
						printf("error: -f requires an argument\n");
						usage();
						exit(0);
					}
					strcpy(filed->conf_file,argv[i]);
					dlog(DEBUG_MAIN,"conf file set to %s",filed->conf_file);
					break;
				default:
						printf("error: unknown option: %c\n",argv[i][1]);
						usage();
						exit(0);
			}
		}
	}
	return;
}

void usage(void) {
	printf("Filed %s\n",VERSION);
	printf("Usage: filed (flags)\n");
	printf("Flags:\n");
	printf("\t-c\tWrite data to the console\n");
	printf("\t-d\tSpecify initial debug level\n");
	printf("\t-h,-?\tThis message\n");
	printf("\t-S\tShow shared memory statistics and exit\n");
	printf("\t-f\tspecify configuration file\n");
	printf("\t-V\tdisplay version info and exit\n");
}

void die(void) {
	/* eventually this will handle clean shutdown of shmem, socket, etc */
	shmem_shutdown();
	dlog(DEBUG_THREAD,"thread closing");
	kill(filed->db_pid,SIGTERM);
	free_parse_hdr(filed->cf);
	unlink("/var/run/filed.pid");
	exit(0);
}

void sighup(int sig) {
	read_cf();
}

void sigterm(int sig) {
	die();
}

void setup_signals(void) {
	signal(SIGTERM,sigterm);
	signal(SIGHUP,sighup);
	signal(SIGINT,sigterm);
	signal(SIGCHLD,SIG_IGN);
}

int get_id(void) {
	int id;
	id = ++filed->id;
	if(id > 65534) filed->id = 0;
	return(id);
}

void version(void) {
	printf("Filed %s\n",VERSION);
	printf("Written by sheer@www.com\n");
	printf("Shouts out to: Greg, Homey Dave, Mark S., Steve \"The man with the food\" Cohen\n");
}

void scan_filesystems(void) {
	/* 
	 * scan for any down filesystems 
	 *
	 */
	struct parse_hdr *p_hdr;
	struct parse *p;
	char filebuf[1024];
	char *c;
	char *d;
	int okay = 0;
	
	p_hdr = filed->cf;
	d = get_one_data(filed->cf,"ping-file");
	if(!d) return;

	while((p_hdr = get_entry(p_hdr,"media-type"))) {
		p = p_hdr->data;
		c = get_token_s(p,3);
		if(c) {
			sprintf(filebuf,"%s/%s",c,d);
			if(!test_open(filebuf)) {
				dlog(DEBUG_FILE,"scan_filesystems(): ping of filesystem %s failed",c);
				okay++;
			}
		}
	
		p_hdr = p_hdr->next;
	}
	
	if(okay) {
		if(!(filed->peer_flags & PFLAG_DOWN)) filed->peer_flags += PFLAG_DOWN;
	} else {
		if((filed->peer_flags & PFLAG_DOWN) && !(filed->peer_flags & PFLAG_ADMINDOWN)) filed->peer_flags -= PFLAG_DOWN;
	}
	return;
}
